#include "Geometry.h"
#include "Render.h"
#include "Image.h"
#include "Mesh.h"

#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <cstring>

#include <stdexcept>

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

string get_suffix(string path) {
	size_t pos = path.rfind(".");
	if (pos == string::npos) {
		return "";
	} else {
		return path.substr(pos);
	}
}

int main(int argc, char *argv[])
{
	string obj_path;
	string mtl_search_dir;
	string output_path = "render.bmp";
	int img_width;
	int img_height;
	double size_factor;
	double pitch = 0;
	double yaw = 0;
	Vec3 light_angle(1,-2,0);
	Color light_color(1,1,1);
	CullMode cullmode = CULL_NONE;
	bool autocompute_normals = false;
	
	if (argc<5)
	{
		cerr << "Usage:" << endl;
		cerr << argv[0] << " <input.obj> <width> <height> <scale-factor> <optional flags...>" << endl;
		cerr << endl;
		cerr << "<input.obj> is the path to a file in the .OBJ file format. RetroRenderer will" << endl
		     << "look for materials (.MTL files) that the .OBJ references in the same directory" << endl
		     << "where it finds the .OBJ." << endl;
		cerr << endl;
		cerr << "<width> and <height> are the size, in pixels, of the output image that you want." << endl
		     << "<scale-factor> is a multiplier that your input model will be scaled by before" << endl
		     << "begin rendered." << endl;
		cerr << endl;
		cerr << "List of optional flags:" << endl;
		cerr << endl;
		cerr << "-o, --output <filename>" << endl;
		cerr << "    Sets the name of the image file that RetroRenderer will write out." << endl
		     << "    RetroRenderer will try to guess the file format from the extension; only" << endl
		     << "    BMP and TGA are supported right now." << endl;
		cerr << endl;
		cerr << "--pitch <degrees>" << endl;
		cerr << "--yaw <degrees>" << endl;
		cerr << "    Rotate the model before rendering it." << endl;
		cerr << endl;
		cerr << "--light-angle <x> <y> <z>" << endl;
		cerr << "    Direction to the light, expressed as a vector. Default is (1, -2, 0)." << endl;
		cerr << endl;
		cerr << "--light-color <r> <g> <b>" << endl;
		cerr << "    Color of the light. Default is (1, 1, 1)." << endl;
		cerr << endl;
		cerr << "--autocompute-normals" << endl;
		cerr << "    If this flag is specified, RetroRenderer will compute face normals that are" << endl;
		cerr << "    perpendicular to the faces and use those instead of the normal vectors" << endl;
		cerr << "    included in the .OBJ file. Useful if your OBJ file has missing or broken" << endl;
		cerr << "    normals." << endl;
		cerr << endl;
		cerr << "--cull <mode>" << endl;
		cerr << "    If this flag is specified, RetroRenderer will skip rendering faces on the" << endl;
		cerr << "    back of the model, as determined by the order of the vertices of the face" << endl;
		cerr << "    (clockwise or counterclockwise). Possible values are 'front', 'back', or" << endl;
		cerr << "    'none'." << endl;
		exit(1);
	}
	
	for (int i=0; i<argc; i++)
	{
		char *arg = argv[i];
		if (i==0)
		{
			continue;
		}
		else if (i==1)
		{
			obj_path = arg;
			
			size_t last_slash_pos = obj_path.find_last_of('/');
			if (last_slash_pos == string::npos) mtl_search_dir = "";
			else mtl_search_dir = obj_path.substr(0, last_slash_pos+1);
		}
		else if (i==2)
		{
			img_width = atoi(arg);
			if (img_width <= 0) { cerr << "bad image width" << endl; exit(1); }
		}
		else if (i==3) 
		{
			img_height = atoi(arg);
			if (img_height <= 0) { cerr << "bad image height" << endl; exit(1); }
		}
		else if (i==4)
		{
			size_factor = atof(arg);
			if (size_factor == 0) { cerr << "bad size factor" << endl; exit(1); }
		}
		else if (string(arg) == "-o" || string(arg) == "--output")
		{
			i++;
			if (i >= argc) { cerr << "--output needs an argument" << endl; exit(1); }
			output_path = argv[i];
		}
		else if (string(arg) == "--pitch")
		{
			i++;
			if (i >= argc) { cerr << "--pitch needs an argument" << endl; exit(1); }
			pitch = atof(argv[i])*M_PI/180;
		}
		else if (string(arg) == "--yaw")
		{
			i++;
			if (i >= argc) { cerr << "--yaw needs an argument" << endl; exit(1); }
			yaw = atof(argv[i])*M_PI/180;
		}
		else if (string(arg) == "--light-angle")
		{
			if (i+3 >= argc) { cerr << "--light-angle needs three argumnts" << endl; exit(1); }
			light_angle.x = atof(argv[i+1]);
			light_angle.y = atof(argv[i+2]);
			light_angle.z = atof(argv[i+3]);
			i += 3;
		}
		else if (string(arg) == "--light-color")
		{
			if (i+3 >= argc) { cerr << "--light-color needs three argumnts" << endl; exit(1); }
			light_color.r = atof(argv[i+1]);
			light_color.g = atof(argv[i+2]);
			light_color.b = atof(argv[i+3]);
			i += 3;
		}
		else if (string(arg) == "--autocompute-normals")
		{
			autocompute_normals = true;
		}
		else if (string(arg) == "--cull")
		{
			i++;
			if (i >= argc) { cerr << "--cull needs an argument" << endl; exit(1); }
			if (string(argv[i]) == "front") cullmode = CULL_FRONT;
			else if (string(argv[i]) == "back") cullmode = CULL_BACK;
			else if (string(argv[i]) == "none") cullmode = CULL_NONE;
			else { cerr << "--cull expects 'front', 'back', or 'none'" << endl; exit(1); }
		}
		else
		{
			cerr << "do not recognize "+string(arg) << endl;
			exit(1);
		}
	}
	
	ifstream model_file(obj_path.c_str(), ios_base::in);
	if (!model_file) {
		cerr << "cannot open file \""+obj_path+"\"" << endl;
		exit(1);
	}
	Mesh model;
	try {
		model = Mesh::from_objfile(model_file, mtl_search_dir);
	} catch (const logic_error &err) {
		cerr << "error parsing file \""+obj_path+"\": "+err.what() << endl;
		exit(1);
	}
	if (model_file.bad()) {
		cerr << "error reading file \""+obj_path+"\"" << endl;
		exit(1);
	}
	model_file.close();
	
	if (model.faces.empty()) {
		cerr << "input file doesn't have any faces" << endl;
		exit(1);
	}
	
	if (autocompute_normals)
	{
		for (list<Face>::iterator it = model.faces.begin();
			it != model.faces.end(); it++)
		{
			Face &f = *it;
			f.autocompute_normals();
		}
	}

	Image canvas(img_width, img_height);
	canvas.clear(Color(0.5,0.5,0.5));
	
	Matrix4 transform = Matrix4::identity;
	
	// Center image in canvas
	transform = transform * Matrix4::translation(Vec3(img_width/2,img_height/2,0));
	
	// Scale image by desired scaling factor
	double sf = size_factor * img_height;
	transform = transform * Matrix4::scaling(Vec3(sf,sf,sf));
	
	// Pitch image
	transform = transform * Matrix4::rotation(pitch, Vec3(1,0,0));
	
	// Yaw image
	transform = transform * Matrix4::rotation(yaw, Vec3(0,1,0));
	
	list<SunLight> lights;
	lights.push_back(SunLight(light_angle, light_color));
	
	cerr << "Rendering..." << endl;
	render(model, transform, lights, canvas, cullmode);
	
	ofstream output_file(output_path.c_str(), ios_base::out);
	string suffix = get_suffix(output_path);
	if (suffix == ".bmp" || suffix == ".BMP") {
		canvas.write_BMP(output_file);
	} else if (suffix == ".tga" || suffix == ".TGA") {
		canvas.write_TGA(output_file);
	} else {
		cerr << "Unrecognized output file suffix '" << suffix << "'; output will be in .BMP format." << endl;
		canvas.write_BMP(output_file);
	}
	output_file.close();
	cerr << "Output successfully written to '" << output_path << "'." << endl;
	
	return 0;
}

